unit isomain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Menus, stypes, ExtCtrls, TeeProcs, TeEngine, Chart;

const
   maxtrans = 35; // Maximum number of transfers with fractionation

type
  stvec = array[1..maxtrans] of double;
  delvec = array[1..maxtrans+maxstate] of double;
  coefarray = array[1..maxtrans,1..maxtrans] of double;
  mat = array[1..sqr(maxtrans)+2*maxtrans] of double;
  namelist = array[1..maxprocess+maxtrans+1] of string[stringlength];
  datavalues = array[1..maxprocess+maxtrans+1] of double;
  TFmIsoMain = class(TForm)
    EdSpecFile: TEdit;
    EdDataFile: TEdit;
    EdOutputFile: TEdit;
    BtnRun: TButton;
    MainMenu1: TMainMenu;
    MIFiles: TMenuItem;
    MIParamFile: TMenuItem;
    MIDriverFile: TMenuItem;
    MIOutputFile: TMenuItem;
    MIBar: TMenuItem;
    MIExit: TMenuItem;
    MIView: TMenuItem;
    MIChart: TMenuItem;
    MITable: TMenuItem;
    BtnClose: TButton;
    DlgOpenSpec: TOpenDialog;
    DlgOpenDataFile: TOpenDialog;
    DlgSaveOutput: TSaveDialog;
    MIAbout: TMenuItem;
    ImgBackground: TImage;
    LblSpecFile: TLabel;
    LblDataFile: TLabel;
    LblOutputFile: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure ChooseSpecFile(Sender: TObject);
    procedure ChooseDataFile(Sender: TObject);
    procedure ChooseOutputFile(Sender: TObject);
    procedure BtnRunClick(Sender: TObject);
    procedure DisplayOutput(Sender: TObject);
    procedure MIExitClick(Sender: TObject);
    procedure InitializeGrid(names:namelist; unt:namelist);
    procedure WritetoGrid(x:datavalues; del:delvec);
    procedure Updateform(Sender: Tobject);
    procedure MIAboutClick(Sender: TObject);
    procedure EdOutputFileKeyPress(Sender: TObject; var Key: Char);
    procedure EdDataFileKeyPress(Sender: TObject; var Key: Char);
    procedure EdSpecFileKeyPress(Sender: TObject; var Key: Char);
  private
    { Private declarations }
    FGridrownumber:integer;
    FRunComplete:Boolean;
    procedure FindA;
    procedure FindL1H1;
    procedure FindaLaH;
    procedure FindLfHf;
    procedure FindL2H2;
    procedure FindPsi2;
    procedure FindDel;
  public
    { Public declarations }
  end;

var
  FmIsoMain: TFmIsoMain;
  ncol, nstat, nflux, nforce:integer;
  ModelDef: Tmodeldef;
  specfilename, datafilename, outfilename:string;
  qbase:double;
  R1,R2,A,phi,
  X1,X2,F1,F2,psi1,psi2,delF1,delF2,L1,L2,H1,H2,LF,HF,Rd:stvec;
  del:delvec;
  colX,colF,delFcol,Fto:array[1..maxtrans] of integer;
  colR,Rfrom,Rto:array[1..sqr(maxtrans)] of integer;
  aL,aH:coefarray;
  time1,time2: double;
  drive:drivearray;
  stat:statearray;
  proc:processarray;
  par:paramarray;

implementation

{$R *.DFM}

uses ifileio, matrix, idisplay, aboutbox;

procedure TFmIsoMain.BtnRunClick(Sender: TObject);
var
   units, names:namelist;
   data1, data2:datavalues;
   i:integer;
begin
 BtnRun.Enabled := False;
 FGridRowNumber := FirstRow;
 for i:= 1 to maxtrans do
    begin
       X1[i] := 0;
       X2[i] := 0;
       F1[i] := 0;
       F2[i] := 0;
       delF1[i] := 0;
       delF2[i] := 0;
       del[i] := 0;
       R1[i] := 0;
       R2[i] := 0;
       phi[i] := 0;
    end;
 for i:= 1 to maxprocess + maxtrans + 1 do
    begin
       data1[i] := 0;
       data2[i] := 0;
       units[i] := '';
       names[i] := '';
    end;

 ReadSpecFile(specfilename);
 // Add code to check if maxstate is exceeded.
 if Ncol > maxprocess+maxtrans+1 then
  Exception.Create('Data array is too small to hold all the data. Delete '
    + 'unneccessary data columns and try again.') ;
    // Equation 5 in Rastetter et. al., 2005, Ecological Applications
 for i:=1 to nstat do psi1[i] := (1 + del[i]/1000.0)*qbase;
 for i:=1 to nflux do
    // Equation 7 in Rastetter et. al., 2005, Ecological Applications
    phi[i] := 1 - qbase*Rd[i]/(psi1[Rfrom[i]]*1000.0);
 OpenDataFile(datafilename, names, units);
 try
   InitializeGrid(names, units);
   ReadDataFile(time1, data1, X1, F1, delF1, R1);
   X2 := X1;
   R2 := R1;
   F2 := F1;
   delF2 := delF1;
   psi2 := psi1;
   time2 := time1;
   data2 := data1;
   WritetoGrid(data1, del);
   while ReadDataFile(time2, data2, X2, F2, delF2, R2) do
    begin
     findA;
     findL1H1;
     findaLaH;
     findLfHf;
     findL2H2;       // Equation 16 in Rastetter et. al., 2005, Ecological Applications
     findpsi2;
     finddel;
     WritetoGrid(data2, del);
  // Set up for next time step.
     X1:=X2;
     R1:=R2;
     F1:=F2;
     delF1:=delF2;
     psi1:=psi2;
     time1:=time2;
    end;
 finally
   CloseDataFile;
   SaveOutput(outfilename);
   FRunComplete := True;
   MiChart.Enabled := True;
   MiTable.Enabled := True;
   BtnRun.Enabled := True;
 end;
end;

procedure TFmIsoMain.findA;
var
   i:integer;
begin
        for i:=1 to nflux do
           begin
               if X1[Rfrom[i]]*X2[Rfrom[i]]<>0 then  // to prevent divide by zero
                  // Equation 2 in Rastetter et. al., 2005, Ecological Applications
                  a[i]:=(time2-time1)*0.5*(R1[i]/X1[Rfrom[i]]+ R2[i]/X2[Rfrom[i]])
               else a[i]:=0;
           end;
end;

procedure TFmIsoMain.findL1H1;
var i:integer;
begin
   for i:= 1 to nstat do
      begin
         // Equation 9 in Rastetter et. al., 2005, Ecological Applications
         L1[i]:=X1[i]/(1+psi1[i]);
         // Equation 8 in Rastetter et. al., 2005, Ecological Applications
         H1[i]:=X1[i]*psi1[i]/(1+psi1[i]);
      end;
end;

procedure TFmIsoMain.findaLaH;
var
   i,j:integer;
   dumL,dumH:double;
begin

   for i:= 1 to nstat do
      begin
         for j:= 1 to nstat do
            begin
              aL[i,j]:=0;
              aH[i,j]:=0;
            end;
      end;
   for i:= 1 to nflux do
      begin
          // Equation 13 in Rastetter et. al., 2005, Ecological Applications
          aL[Rto[i],Rfrom[i]]:=aL[Rto[i],Rfrom[i]]+
               a[i]*(1+psi1[Rfrom[i]])/(1+phi[i]*psi1[Rfrom[i]]);
          // Equation 12 in Rastetter et. al., 2005, Ecological Applications
          aH[Rto[i],Rfrom[i]]:=aH[Rto[i],Rfrom[i]]+
               a[i]*phi[i]*(1+psi1[Rfrom[i]])/(1+phi[i]*psi1[Rfrom[i]]);
      end;
   for j:=1 to nstat do
      begin
         dumL:=0;
         dumH:=0;
         for i:= 1 to nstat do
            begin
               dumL:=dumL - aL[i,j];
               dumH:=dumH - aH[i,j];
            end;
         aL[j,j]:=dumL; 
         aH[j,j]:=dumH;
      end;
end;

procedure TFmIsoMain.findLfHf;
var
   i:integer;
   psi:double;
begin
   for i:= 1 to nstat do
      begin
         // Equation 6 in Rastetter et. al., 2005, Ecological Applications
         psi:=(1+delF1[i]/1000.0)*qbase;
         // Equation 11 in Rastetter et. al., 2005, Ecological Applications
         Lf[i]:=(time2-time1)*F1[i]/(1+psi);
         // Equation 10 in Rastetter et. al., 2005, Ecological Applications
         Hf[i]:=(time2-time1)*F1[i]*psi/(1+psi);
      end;
end;

procedure TFmIsoMain.findL2H2;
var
   dum:double;
   nit,i:integer;
   L1a,H1a:stvec;
begin
   dum:=aL[1,1];
   for i:= 1 to nstat do if dum>aL[i,i] then dum:=aL[i,i]; {remember the a[i,i] are negative}
   for i:= 1 to nstat do if dum>aH[i,i] then dum:=aH[i,i];
   dum:= -10*dum*(time2-time1)+1;
   nit:= trunc(dum);
   if nit< 10 then nit:=10;

   L1a:=L1;
   H1a:=H1;
   //  This loop calculates Equation 16 in Rastetter et. al., 2005, Ecological Applications
   for i:= 1 to nit do
      begin
         mult(aL,L1a,L2,maxtrans,maxtrans,1);
         Madd(LF,L2,L2,maxtrans,1);
         Smult(L2,L2,(time2-time1)/nit,maxtrans,1);
         Madd(L1a,L2,L2,maxtrans,1);
         mult(aH,H1a,H2,maxtrans,maxtrans,1);
         Madd(HF,H2,H2,maxtrans,1);  
         Smult(H2,H2,(time2-time1)/nit,maxtrans,1);
         Madd(H1a,H2,H2,maxtrans,1);
         L1a:=L2;
         H1a:=H2;
      end;
end;

procedure TFmIsoMain.findpsi2;
var i:integer;
begin
   // Equation 18 in Rastetter et. al., 2005, Ecological Applications
   for i:= 1 to nstat do psi2[i]:=H2[i]/L2[i];
end;

procedure TFmIsoMain.finddel;
var
 i,j,k:integer;
 aLdum, aHdum, dum:double;
begin
 // Equation 19 in Rastetter et. al., 2005, Ecological Applications
 for i:= 1 to nstat do del[i]:= (psi2[i]/qbase - 1)*1000.0;
 for j := ncol + nstat to ncol + nstat + nflux - 1 do
   begin
     k := j - ncol - nstat + 1;
     // Equation 13 in Rastetter et. al., 2005, Ecological Applications
     aLdum := a[k]*(1+psi1[Rfrom[k]])/
                (1+phi[k]*psi1[Rfrom[k]]);
     // Equation 12 in Rastetter et. al., 2005, Ecological Applications
     aHdum := a[k]*phi[k]*(1+psi1[Rfrom[k]])/
                (1+phi[k]*psi1[Rfrom[k]]);
     dum := qbase*aLdum*L1[Rfrom[k]];
     // Equation 3 in Rastetter et. al., 2005, Ecological Applications
     if dum > 0 then dum := (aHdum*H1[Rfrom[k]]/dum - 1)*1000 else dum:=0;
     del[k+nstat] := dum;
   end;
end;

procedure TFmIsoMain.FormCreate(Sender: TObject);
begin
 ForceCurrentDirectory := True;
 ModelDef.modelname := 'NESIS';
 ModelDef.versionnumber := '1.3.2';
 ModelDef.timeunit := '';
 ModelDef.numdrive := 0;
 ModelDef.numstate := 0;
 ModelDef.numprocess := 0;
 ModelDef.numparam := 0;
 FRunComplete := False;
end;

procedure TFmIsoMain.ChooseSpecFile(Sender: TObject);
var
 ThisEdit:Tedit;
 oldfilename:string;
begin
 oldfilename := specfilename;
// If the user typed directly in the edit box
 if Sender is TEdit then
  begin
    ThisEdit := Sender as Tedit;
    Specfilename:=ThisEdit.text  // Set the specfile
  end
 else if Sender is Tlabel then// User used the menu or clicked on the Spec label
  begin   // Show the open file dialog
   ThisEdit := ((Sender as TLabel).Focuscontrol as TEdit);
    // First set the default filename to the current specfilename
   DlgOpenSpec.filename := specfilename;
    // Show the dialog and if the user clicks OK, set the new spec filename
   if DlgOpenspec.execute then
    begin
      specfilename := DlgOpenspec.filename;
      ThisEdit.Text := specfilename;
    end;
  end
 else  // Sender is main menu
  begin
   DlgOpenSpec.filename := specfilename;
    // Show the dialog and if the user clicks OK, set the new driver filename
   if DlgOpenSpec.execute then
    begin
      specfilename := DlgOpenSpec.filename;
      EdSpecFile.Text := specfilename;
    end;
  end;
 if specfilename <> oldfilename then FRunComplete := False;
 Updateform(EdSpecFile);
end;

procedure TFmIsoMain.ChooseDataFile(Sender: TObject);
var
 ThisEdit:Tedit;
 oldfilename:string;
begin
 oldfilename := datafilename;
// If the user typed directly in the edit box
 if Sender is TEdit then
  begin
    ThisEdit := Sender as Tedit;
    Datafilename:=ThisEdit.text  // Set the datafile
  end
 else if Sender is Tlabel then// User used the menu or clicked on the data label
  begin   // Show the open file dialog
   ThisEdit := ((Sender as TLabel).Focuscontrol as TEdit);
    // First set the default filename to the current data file
   DlgOpenDataFile.filename := datafilename;
    // Show the dialog and if the user clicks OK, set the new data filename
   if DlgOpenDataFile.execute then
    begin
      datafilename := DlgOpenDataFile.filename;
      ThisEdit.Text := datafilename;
    end;
  end
 else  // Sender is main menu
  begin
   DlgOpenDataFile.filename := datafilename;
    // Show the dialog and if the user clicks OK, set the new driver filename
   if DlgOpenDataFile.execute then
    begin
      datafilename := DlgOpenDataFile.filename;
      EdDataFile.Text := datafilename;
    end;
  end;
 if datafilename <> oldfilename then FRunComplete := False;
 Updateform(EdDataFile);
end;

procedure TFmIsoMain.ChooseOutputFile(Sender: TObject);
var
 ThisEdit:Tedit;
begin
 if Sender is TEdit then
  begin
    ThisEdit := Sender as Tedit;
    Outfilename := ThisEdit.text  // Set the output file
  end
 else if Sender is Tlabel then  // User used the menu or clicked on the Output label
  begin  // Show the Save file dialog
    // First set the default filename to the current output file.
   DlgSaveOutput.filename := outfilename;
   // Show the dialog and if the user chooses OK, set the outfilename to the new file.
   if DlgSaveOutput.execute then
    begin
      outfilename := DlgSaveOutput.filename;
      EdOutputFile.Text := outfilename
    end;
  end
 else  // Sender is main menu
  begin
   DlgSaveOutput.filename := outfilename;
    // Show the dialog and if the user clicks OK, set the new driver filename
   if DlgSaveOutput.execute then
    begin
      outfilename := DlgSaveOutput.filename;
      EdOutputFile.Text := outfilename;
    end;
  end;
 Updateform(EdOutputFile);
end;

procedure TFmIsoMain.MIExitClick(Sender: TObject);
begin
 FmIsoMain.Close;
end;

{ This procedure shows the output from the model run in table or chart format.
  It uses the form defined in display.pas. }
procedure TFmIsoMain.DisplayOutput(Sender: TObject);
begin
 if (Sender as TMenuItem).Name = 'MITable' then
  FmDisplayData.DisplayStyle := dsTable // Set the type of display.
 else
  FmDisplayData.DisplayStyle := dsChart;
 FmDisplayData.showmodal; // Show the Display Form.
end;

procedure TFmIsoMain.InitializeGrid(names:namelist; unt:namelist);
var
 i:integer;
begin
  FmDisplayData.ClearGrid;
  FmDisplayData.chOutput.Title.Text.text := 'Isotope Model Output'; // Chart title
  with FmDisplayData.SgModelOutput1 do // Set up the Stringgrid
   begin
    colcount := ncol + nstat + nflux;
    for i := 0 to ncol + nstat + nflux - 1 do
      begin
        cells[i,0] := inttostr(i+1);
        cells[i,1] := names[i+1];
        cells[i,2] := unt[i+1];
      end;
   end;
 FmDisplayData.NeedtoReadData := False;

 FmDisplayData.LbxSeries.Items.Clear;
 for i := 1 to ncol + nstat + nflux do
                FmDisplayData.LbxSeries.Items.Add(names[i]);
end;

procedure TFmIsoMain.WritetoGrid(x:datavalues; del:delvec);
var
 j:integer;
begin
 FmDisplayData.Caption := 'Model Output - File: ' + OutFilename;
 FmDisplayData.SgModelOutput1.rowcount :=
                 FmDisplayData.SgModelOutput1.rowcount + 1;
 for j := 0 to ncol - 1 do
   FmDisplayData.SgModelOutput1.Floats[j,FGridRowNumber] :=
        x[j+1];
 for j := ncol to ncol + nstat - 1 do
   FmDisplayData.SgModelOutput1.Floats[j,FGridRowNumber] :=
        del[j-ncol+1];
 for j := ncol + nstat to ncol + nstat + nflux - 1 do
     FmDisplayData.SgModelOutput1.Floats[j,FGridRowNumber] := del[j-ncol+1];
 FGridrownumber := FGridrownumber + 1;      // Increment row number writing to
end;

procedure TFmIsoMain.Updateform(Sender:Tobject);
begin
 if (specfilename <> '') and (datafilename <> '') and (outfilename <> '') then
   BtnRun.enabled := true
 else
   BtnRun.enabled := false;
 if not FRunComplete then
   begin
     MiChart.Enabled := False;
     MiTable.Enabled := False;
   end;
end;

procedure TFmIsoMain.MIAboutClick(Sender: TObject);
begin
 FmAbout.ShowModal;
end;

procedure TFmIsoMain.EdOutputFileKeyPress(Sender: TObject; var Key: Char);
begin
  if (Key = Chr(13)) then ChooseOutputFile(Sender);
end;

procedure TFmIsoMain.EdDataFileKeyPress(Sender: TObject; var Key: Char);
begin
  if (Key = Chr(13)) then ChooseDataFile(Sender);
end;

procedure TFmIsoMain.EdSpecFileKeyPress(Sender: TObject; var Key: Char);
begin
  if (Key = Chr(13)) then ChooseSpecFile(Sender);
end;

end.


