unit matrix;

interface

uses isomain;

procedure Mprint(Var X;m,n:integer);
procedure Madd(var X,Y,Z;m,n:integer);
Procedure SMult(Var X,Y; s:double; m,n:integer);
Procedure Mult(Var X,Y,Z;l,m,n:integer);
Procedure Trans(Var X,Y; m,n:integer);
Procedure INV(Var Z,X;n:integer);
Procedure Mexp(Var X,Y; n:integer);
Procedure Check(Var X; m,n:integer);


implementation

procedure Mprint(Var X;m,n:integer);
{print matrix X[m,n] to default devise}
//mat is a vector long enough to hold all elements of array X
// the absolute structure makes this procedure work for arrays of any dimension
// I think Var declaration is needed to allow the absolute declaration
// may need to change this to an open array declaration in Delphi, i.e.,  Var X: array of double; ....
// these procedures expect an output file called outfile to be open

Var
  X1: mat absolute X;

Begin
 { for i:=1 to m do
    begin
      for j:= 1 to n do
        begin
          r:=(i-1)*n+j;
          write(outfile,X1[r]:10,' ');
        end;
      writeln(outfile,' ');
    end;
writeln(outfile,' ');
writeln(outfile,' ');  }
End;

procedure Madd(var X,Y,Z;m,n:integer);
{Z[m,n] = X[m,n] + Y[m,n]}
{m rows n columns}
Var
  X1: mat absolute X;
  Y1: mat absolute Y;
  Z1: mat absolute Z;
  i,j,k:integer;

Begin
  for i:= 1 to m do
    begin
      for j:= 1 to n do
        begin
          k:=(i-1)*n+j;
          Z1[k]:= X1[k] + Y1[k];
        end;
    end;
End;

Procedure SMult(Var X,Y; s:double; m,n:integer);
{Scalar multiply Y[m,n] = s*X[m,n]}
Var
  X1: mat absolute X;
  Y1: mat absolute Y;
  i,j,k: integer;

Begin
  for i:= 1 to m do
    begin
      for j:= 1 to n do
        begin
          k:=(i-1)*n+j;
          Y1[k] := s*X1[k];
        end;
    end;
End;

Procedure Mult(Var X,Y,Z;l,m,n:integer);
{Matrix multiply Z[l,n] = X[l,m] x Y[m,n]}
{ X has l rows and m columns
  Y has m rows and n columns
  Z has l rows and n columns}
Var
  X1: mat absolute X;
  Y1: mat absolute Y;
  Z1: mat absolute Z;
  dum:mat;
  i,j,k,r,s,t:integer;

Begin
  for i:= 1 to l do
    begin
      for j:= 1 to n do
        begin
          r:=(i-1)*n+j;
          dum[r]:=0;
        end;
    end;
  for i:= 1 to l do
    begin
      for j:= 1 to n do
        begin
          for k:= 1 to m do
            begin
              r:=(i-1)*n+j;
              s:=(i-1)*m+k;
              t:=(k-1)*n+j;
              dum[r]:= dum[r] + X1[s]*Y1[t];
            end;
        end;
    end;
  for i:=1 to l do
    begin
      for j:= 1 to n do
        begin
          r:=(i-1)*n+j;
          Z1[r]:=dum[r];
        end;
    end;
End;

Procedure Trans(Var X,Y; m,n:integer);
{Transpose Y[n,m] = X[m,n]^T}

Var
  X1: mat absolute X;
  Y1: mat absolute Y;
  i,j,r,s:integer;

Begin
  for i:= 1 to m do
    begin
      for j:= 1 to n do
        begin
          r:=(i-1)*n+j;
          s:=(j-1)*m+i;
          Y1[r]:= X1[s];
        end;
    end;
End;

Procedure INV(Var Z,X;n:integer);
{invert X[n,n] = Z[n,n]^-1}
Label 10,20,30;

Var
  X1: mat absolute X;
  Y1: mat;
  Z1: mat absolute Z;
  i,j,k,r,s:integer;
  dum:double;

Begin
  for i:= 1 to n do
    begin
      for j:= 1 to n do
        begin
          r:=(i-1)*n+j;
          if i=j then X1[r]:=1 else X1[r]:=0;
          Y1[r]:=Z1[r];
        end;
    end;
  for i:= 1 to n do
    begin
      r:=(i-1)*n+i;
      dum:=Y1[r];
      if (dum=0) and (i<>n) then
        begin
{          writeln(outfile,'zero diagonal');      }
          k:=0;
          s:=i*n+i;
          while Y1[s]= 0 do
            begin
              k:=k+1;
              if i+k>=n then goto 20;
              s:=(i+k)*n+i;
            end;

          for j:= 1 to n do
            begin
              r:=(i-1)*n+j;
              s:=(i+k)*n+j;
              Y1[r]:=Y1[r]+Y1[s];
              X1[r]:=X1[r]+X1[s];
            end;
          r:=(i-1)*n+i;
          dum:=Y1[r];
        end;
      if (dum=0) and (i=n) then  goto 20;
      for j:= 1 to n do
        begin
          r:=(i-1)*n+j;
          X1[r]:=X1[r]/dum;
          Y1[r]:=Y1[r]/dum;
        end;
      for k:=1 to n do
        begin
          if k=i then goto 10;
          r:=(k-1)*n+i;
          dum:=Y1[r];
          for j:= 1 to n do
            begin
              r:=(k-1)*n+j;
              s:=(i-1)*n+j;
              Y1[r]:=Y1[r]-Y1[s]*dum;
              X1[r]:=X1[r]-X1[s]*dum;
            end;
     10:end;
    end;
goto 30;
20: //writeln(outfile,'singular matrix');
   for i:= 1 to n do
     begin
       for j:=1 to n do
         begin
           r:=(i-1)*n+j;
           X1[r]:=0;
         end;
      end;
30:End;

Procedure Mexp(Var X,Y; n:integer);
{Matrix exponential Y[n,n] = exp(X[n,n])}
Var
  X1: mat absolute X;
  Y1: mat absolute Y;
  D1,D2,D3: mat;
  i,j,k,r:integer;
  s,s1: double;

function fact(i:double):double;
var m:double;
begin
  m:=1;
  while i>1 do
    begin
      m:=m*i;
      i:=i-1;
    end;
  fact:=m;
end;

Begin
  for i:= 1 to n do
    begin
      for j:= 1 to n do
        begin
          r:=(i-1)*n+j;
          if i=j then Y1[r]:=1 else Y1[r]:=0;
        end;
    end;
  D1:=X1;
  D3:=X1;
  s1:=1;
  s:=1;
  while s>0.000001 do
    begin
      Madd(Y1,D1,Y1,n,n);
      s1:=s1+1;
      mult(D3,X1,D2,n,n,n);
      D3:=D2;
      s:=fact(s1);
      s:=1/s;
      Smult(D2,D1,s,n,n);
      S:=0;
      for j:=1 to n do
        begin
          for k:= 1 to n do
            begin
              r:=(j-1)*n+k;
              s:=s+abs(D1[r]);
            end;
        end;
    end;
  Madd(Y1,D1,Y1,n,n);
End;

Procedure Check(Var X; m,n:integer);
{set all matrix elements < 1.0e-10 to 0}
Var
   X1: mat absolute X;
   i, j, r:integer;
Begin
   For i:=1 to m do
      For j:=1 to n do
      begin
         r:=(i-1)*n + j;
         If abs(X1[r]) < 1.0E-10
            Then   X1[r]:=0;
      end;
End;

end.